/*
 * Copyright 1995, 1996 by Jun Yei.			All Rights Reserved
 *
 * Permission to retain, use, modify, copy, and distribute ChinesePower
 * in source or binary and its documentation (hereafter, the Software)
 * for noncommercial purpose is hereby granted to you without a fee,
 * provided that this entire copyright and permission notice appear in
 * all such copies, that no charge be associated with such copies,
 * that distribution of derivative works (including value-added
 * distributions such as with additional input dictionaries or fonts)
 * include clarification that such added or derived parts are not from
 * the original Software, and that the names of the author(s) not be
 * used to endorse or promote such works. The author(s) of the software
 * makes no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 * THE AUTHOR(S) OF THE SOFTWARE DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL THE AUTHOR(S) OF THE SOFTWARE BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Jun Yei, Ottawa, ON, Canada
 */

#include "chpower.h"

#define MAX_GHANDS 16   /* maximum # of GRAF handles */

byte           fsgamcr[256];

XImage *CreateXImage();
XImage *Pic8ToXImage();
XImage *Pic24ToXImage();

/*********************/
void InitSpline(x,y,n,y2)
     int *x, *y, n;
     double *y2;
{
  /* given arrays of data points x[0..n-1] and y[0..n-1], computes the
     values of the second derivative at each of the data points
     y2[0..n-1] for use in the splint function */
 
  int i,k;
  double p,qn,sig,un,u[MAX_GHANDS];
 
  y2[0] = u[0] = 0.0;
 
  for (i=1; i<n-1; i++) {
    sig = ((double) x[i]-x[i-1]) / ((double) x[i+1] - x[i-1]);
    p = sig * y2[i-1] + 2.0;
    y2[i] = (sig-1.0) / p;
    u[i] = (((double) y[i+1]-y[i]) / (x[i+1]-x[i])) -
           (((double) y[i]-y[i-1]) / (x[i]-x[i-1]));
    u[i] = (6.0 * u[i]/(x[i+1]-x[i-1]) - sig*u[i-1]) / p;
  }
  qn = un = 0.0;
 
  y2[n-1] = (un-qn*u[n-2]) / (qn*y2[n-2]+1.0);
  for (k=n-2; k>=0; k--)
    y2[k] = y2[k]*y2[k+1]+u[k];
}
 
 
/*********************/
double EvalSpline(xa,ya,y2a,n,x)
double y2a[],x;
int n,xa[],ya[];
{
  int klo,khi,k;
  double h,b,a;
 
  klo = 0;
  khi = n-1;
  while (khi-klo > 1) {
    k = (khi+klo) >> 1;
    if (xa[k] > x) khi = k;
    else klo = k;
  }
  h = xa[khi] - xa[klo];
  if (h==0.0) h = 1.0; /* FatalError("bad xvalues in splint\n")*/
  a = (xa[khi]-x)/h;
  b = (x-xa[klo])/h;
  return (a*ya[klo] + b*ya[khi] + ((a*a*a-a)*y2a[klo] +(b*b*b-b)*y2a[khi])
          * (h*h) / 6.0);
}
 

/*********************/
void GenerateFSGamma()
{
  /* this function generates the Floyd-Steinberg gamma curve (fsgamcr)
 
     This function generates a 4 point spline curve to be used as a
     non-linear grey 'colormap'.  Two of the points are nailed down at 0,0
     and 255,255, and can't be changed.  You specify the other two.  If
     you specify points on the line (0,0 - 255,255), you'll get the normal
     linear reponse curve.  If you specify points of 50,0 and 200,255, you'll
     get grey values of 0-50 to map to black (0), and grey values of 200-255
     to map to white (255) (roughly).  Values between 50 and 200 will cover
     the output range 0-255.  The reponse curve will be slightly 's' shaped. */
 
  int i,j;
  static int x[4] = {0,16,240,255};
  static int y[4] = {0, 0,255,255};
  double yf[4];
 
  InitSpline(x, y, 4, yf);
 
  for (i=0; i<256; i++) {
    j = (int) EvalSpline(x, y, yf, 4, (double) i);
    if (j<0) j=0;
    else if (j>255) j=255;
    fsgamcr[i] = j;
  }
}

/***************************************************/
void xvbzero(s, len)
     char   *s;
     size_t  len;
{
  for ( ; len>0; len--) *s++ = 0;
}

/********************************/
int ReadFileType(fname)
     char *fname;
{
  /* opens fname (which *better* be an actual file by this point!) and
     reads the first couple o' bytes.  Figures out what the file's likely
     to be, and returns the appropriate RFT_*** code */


  FILE *fp;
  byte  magicno[30];    /* first 30 bytes of file */
  int   rv, n;

  if (!fname) return RFT_ERROR;   /* shouldn't happen */

  fp = My_path_open(getenv(PIC_FILE_PATH), fname, "r");
  if (!fp) return RFT_ERROR;

  n = fread(magicno, (size_t) 1, (size_t) 30, fp);  
  fclose(fp);

  if (n<30) return RFT_UNKNOWN;    /* files less than 30 bytes long... */

  rv = RFT_UNKNOWN;

  if (strncmp((char *) magicno,"GIF87a", (size_t) 6)==0 ||
      strncmp((char *) magicno,"GIF89a", (size_t) 6)==0)        rv = RFT_GIF;

  else if (strncmp((char *) magicno,"VIEW", (size_t) 4)==0 ||
	   strncmp((char *) magicno,"WEIV", (size_t) 4)==0)     rv = RFT_PM;

  else if (magicno[0] == 'P' && magicno[1]>='1' && 
	   magicno[1]<='6')                             rv = RFT_PBM;

  /* note: have to check XPM before XBM, as first 2 chars are the same */
  else if (strncmp((char *) magicno, "/* XPM */", (size_t) 9)==0) rv = RFT_XPM;

  else if (strncmp((char *) magicno,"#define", (size_t) 7)==0 ||
	   (magicno[0] == '/' && magicno[1] == '*'))    rv = RFT_XBM;

  else if (magicno[0]==0x59 && (magicno[1]&0x7f)==0x26 &&
	   magicno[2]==0x6a && (magicno[3]&0x7f)==0x15) rv = RFT_SUNRAS;

  else if (magicno[0] == 'B' && magicno[1] == 'M')      rv = RFT_BMP;

  else if (magicno[0]==0x52 && magicno[1]==0xcc)        rv = RFT_UTAHRLE;

  else if ((magicno[0]==0x01 && magicno[1]==0xda) ||
	   (magicno[0]==0xda && magicno[1]==0x01))      rv = RFT_IRIS;

/*  else if (magicno[0]==0x1f && magicno[1]==0x9d)        rv = RFT_COMPRESS;

#ifdef GUNZIP
  else if (magicno[0]==0x1f && magicno[1]==0x8b)        rv = RFT_COMPRESS;
#endif
*/
  else if (magicno[0]==0x0a && magicno[1] <= 5)         rv = RFT_PCX;

  else if (strncmp((char *) magicno,   "FORM", (size_t) 4)==0 && 
	   strncmp((char *) magicno+8, "ILBM", (size_t) 4)==0)   rv = RFT_IFF;

  else if (magicno[0]==0 && magicno[1]==0 &&
	   magicno[2]==2 && magicno[3]==0 &&
	   magicno[4]==0 && magicno[5]==0 &&
	   magicno[6]==0 && magicno[7]==0)              rv = RFT_TARGA;

  else if (magicno[4]==0x00 && magicno[5]==0x00 &&
	   magicno[6]==0x00 && magicno[7]==0x07)        rv = RFT_XWD;

  else if (strncmp((char *) magicno,"SIMPLE  ", (size_t) 8)==0 && 
	   magicno[29] == 'T')                          rv = RFT_FITS;
  

#ifdef HAVE_JPEG
  else if (magicno[0]==0xff && magicno[1]==0xd8 && 
	   magicno[2]==0xff)                            rv = RFT_JFIF;
#endif

#ifdef HAVE_TIFF
  else if ((magicno[0]=='M' && magicno[1]=='M') ||
	   (magicno[0]=='I' && magicno[1]=='I'))        rv = RFT_TIFF;
#endif

#ifdef HAVE_PDS
  else if (strncmp((char *) magicno,  "NJPL1I00", (size_t) 8)==0 ||
	   strncmp((char *) magicno+2,"NJPL1I",   (size_t) 6)==0 ||
           strncmp((char *) magicno,  "CCSD3ZF",  (size_t) 7)==0 ||
	   strncmp((char *) magicno+2,"CCSD3Z",   (size_t) 6)==0 ||
	   strncmp((char *) magicno,  "LBLSIZE=", (size_t) 8)==0)
      rv = RFT_PDSVICAR;
#endif

#ifdef GS_PATH
  else if (strncmp((char *) magicno, "%!",     (size_t) 2)==0 ||
	   strncmp((char *) magicno, "\004%!", (size_t) 3)==0)   rv = RFT_PS;
#endif

  return rv;
}

/********************************/
int ReadPicFile(fname, ftype, pinfo, quick)
     char    *fname;
     int     ftype, quick;
     GRaphe  *pinfo;
{
  /* if quick is set, we're being called to generate icons, or something
     like that.  We should load the image as quickly as possible.  Currently,
     this only affects the LoadPS routine, which, if quick is set, only
     generates the page file for the first page of the document */

  int rv = 0;

  /* by default, most formats aren't multi-page */

  switch (ftype) {
  case RFT_GIF:     rv = LoadGIF   (fname, pinfo);         break;
  case RFT_PM:      rv = LoadPM    (fname, pinfo);         break;
/*  case RFT_PBM:     rv = LoadPBM   (fname, pinfo);         break;
 */ case RFT_XBM:     rv = LoadXBM   (fname, pinfo);         break;
/*  case RFT_SUNRAS:  rv = LoadSunRas(fname, pinfo);         break;
*/  case RFT_BMP:     rv = LoadBMP   (fname, pinfo);         break;
/*  case RFT_UTAHRLE: rv = LoadRLE   (fname, pinfo);         break;
  case RFT_IRIS:    rv = LoadIRIS  (fname, pinfo);         break;
*/  case RFT_PCX:     rv = LoadPCX   (fname, pinfo);         break;
/*  case RFT_IFF:     rv = LoadIFF   (fname, pinfo);         break;
  case RFT_TARGA:   rv = LoadTarga (fname, pinfo);         break;
  case RFT_XPM:     rv = LoadXPM   (fname, pinfo);         break;
  case RFT_XWD:     rv = LoadXWD   (fname, pinfo);         break;
  case RFT_FITS:    rv = LoadFITS  (fname, pinfo, quick);  break;

#ifdef HAVE_JPEG
  case RFT_JFIF:    rv = LoadJFIF  (fname, pinfo, quick);    break;
#endif

#ifdef HAVE_TIFF
  case RFT_TIFF:    rv = LoadTIFF  (fname, pinfo);           break;
#endif

#ifdef HAVE_PDS
  case RFT_PDSVICAR: rv = LoadPDS  (fname, pinfo);           break;
#endif

#ifdef GS_PATH
  case RFT_PS:      rv = LoadPS    (fname, pinfo, quick);    break;
#endif
*/
  }
  return rv;
}

/***********************************/
void graphicsZoom(viewport, graph, newZoom)
VIewportdata *viewport;
GRaphe *graph;
int newZoom;
{
	/* if same size, and Ximage created, do nothing */
	if (graph->zoomNum == newZoom) return;

	graph->zoomNum = newZoom;

	if(graph->Image)
	{
		free(graph->Image->data);
		XFree(graph->Image);
		graph->Image = NULL;
	}

	graph->curW = XmConvertUnits(viewport->resource.WorkArea, XmHORIZONTAL, Xm100TH_POINTS, ( graph->w * newZoom / 100 ) * PICSCALE, XmPIXELS);
	graph->curH = XmConvertUnits(viewport->resource.WorkArea, XmVERTICAL, Xm100TH_POINTS, ( graph->h * newZoom / 100 ) * PICSCALE, XmPIXELS);

	GenerateZoomedPic(viewport->resource.dpy, DefaultVisual(viewport->resource.dpy, DefaultScreen(viewport->resource.dpy)), graph);
}

/***********************************/
void GenerateZoomedPic(theDisp, theVisual, graph)
Display *theDisp;
Visual *theVisual;
GRaphe *graph;
{
	int     cy,ex,ey,*cxarr, *cxarrp;
	byte    *epic, *clptr,*elptr,*epptr;
	int	cWIDE = graph->w, cHIGH = graph->h;
	int	eWIDE = graph->curW, eHIGH = graph->curH;
	int bperpix;

	clptr = NULL;  cxarrp = NULL;  cy = 0;  /* shut up compiler */

	/* run the rescaling algorithm */

	bperpix = (graph->picType == PIC8) ? 1 : 3;

	/* create a new epic of the appropriate size */

	epic = (byte *) malloc((size_t) (eWIDE * eHIGH * bperpix));

/*	if (!epic) FatalError("GenerateEpic():  unable to malloc 'epic'");
*/
	/* the scaling routine.  not really all that scary after all... */

	/* OPTIMIZATON:  Malloc an eWIDE array of ints which will hold the
	   values of the equation px = (pWIDE * ex) / eWIDE.  Faster than doing 
	   a mul and a div for every point in picture */

	cxarr = (int *) malloc(eWIDE * sizeof(int));

/*	if (!cxarr) FatalError("unable to allocate cxarr");
*/
	for (ex=0; ex<eWIDE; ex++) 
		cxarr[ex] = bperpix * ((cWIDE * ex) / eWIDE);

	elptr = epptr = epic;

	for (ey=0;  ey<eHIGH;  ey++, elptr+=(eWIDE*bperpix))
	{
		cy = (cHIGH * ey) / eHIGH;
		epptr = elptr;
		clptr = graph->pic + (cy * cWIDE * bperpix);

		if (bperpix == 1)
		{
			for (ex=0, cxarrp = cxarr;  ex < eWIDE;  ex++, epptr++) 
				*epptr = clptr[*cxarrp++];
		}
		else
		{
			int j;  byte *cp;

			for (ex=0, cxarrp = cxarr; ex<eWIDE; ex++,cxarrp++)
			{
				cp = clptr + *cxarrp;
				for (j=0; j<bperpix; j++) 
	    				*epptr++ = *cp++;
			}
      		}
    	}

	free(cxarr);

	graph->Image = CreateXImage(theDisp, theVisual, epic, graph->picType, eWIDE, eHIGH, graph->r, graph->g, graph->b);

	free(epic);
}

/************************/
static void floydDitherize(ximage,pic824,ptype, wide, high, rmap, gmap, bmap)
     XImage *ximage;
     byte   *pic824, *rmap, *gmap, *bmap;
     int     ptype, wide, high;
{
	/* does floyd-steinberg ditherizing algorithm.  
	 *
	 * takes a wide*high input image, of type 'ptype' (PIC8, PIC24)
	 *     (if PIC8, colormap is specified by rmap,gmap,bmap)
	 *
	 * output is a 1-bit per pixel XYBitmap, packed 8 pixels per byte
	 *
	 * Note: this algorithm is *only* used when running on a 1-bit display
	 */
  
	register byte   pix8, bit;
	int            *thisline, *nextline;
	int            *thisptr, *nextptr, *tmpptr;
	int             i, j, err, bperpix, bperln, order;
	byte           *pp, *image, w1, b1, w8, b8, rgb[256];
  
  
	if (ptype == PIC8)
	{   /* monoify colormap */
		for (i=0; i<256; i++)
			rgb[i] = MONO(rmap[i], gmap[i], bmap[i]);
	}


	image   = (byte *) ximage->data;
	bperln  = ximage->bytes_per_line;
	order   = ximage->bitmap_bit_order;
	bperpix = (ptype == PIC8) ? 1 : 3;


	thisline = (int *) malloc(wide * sizeof(int));
	nextline = (int *) malloc(wide * sizeof(int));
/*	if (!thisline || !nextline) 
		FatalError("ran out of memory in floydDitherize1()\n");
*/

	/* load up first line of picture */
	pp = pic824;

	if (ptype == PIC24)
	{
		for (j=0, tmpptr = nextline; j<wide; j++, pp+=3)
			*tmpptr++ = fsgamcr[MONO(pp[0], pp[1], pp[2])];
	}
	else
	{
		for (j=0, tmpptr = nextline; j<wide; j++, pp++)
		*tmpptr++ = fsgamcr[rgb[*pp]];
	}

      
	w1 = 0x1;  b1= 0x0;
	w8 = w1<<7;  b8 = b1<<7;        /* b/w bit in high bit */
  

	for (i=0; i<high; i++)
	{
		/* get next line of image */
		tmpptr = thisline;  thisline = nextline;  nextline = tmpptr;  /* swap */

		if (i!=high-1)
		{
			pp = pic824 + (i+1) * wide * bperpix;

			if (ptype == PIC24)
			{
				for (j=0, tmpptr = nextline; j<wide; j++, pp+=3)
					*tmpptr++ = fsgamcr[MONO(pp[0], pp[1], pp[2])];
      			}
			else
			{
				for (j=0, tmpptr = nextline; j<wide; j++, pp++)
					*tmpptr++ = fsgamcr[rgb[*pp]];
			}
		}

		thisptr = thisline;  nextptr = nextline;

		pp  = image + i*bperln;

		if (order==LSBFirst)
		{
			bit = pix8 = 0;

			for (j=0; j<wide; j++, thisptr++, nextptr++)
			{
				if (*thisptr<128) { err = *thisptr;     pix8 |= b8; }
	             		else { err = *thisptr-255; pix8 |= w8; }

				if (bit==7) { *pp++ = pix8;  bit=pix8=0; }
	       			else { pix8 >>= 1;  bit++; }

				if (j<wide-1) thisptr[1] += ((err*7)/16);

				if (i<high-1)
				{
	  				nextptr[0] += ((err*5)/16);
	  				if (j>0)      nextptr[-1] += ((err*3)/16);
	  				if (j<wide-1) nextptr[ 1] += (err/16);
				}
      			}

      			if (bit) *pp++ = pix8>>(7-bit);  /* write partial byte at end of line */
    		}
    		else
		{   /* order==MSBFirst */
			bit = pix8 = 0;
			for (j=0; j<wide; j++, thisptr++, nextptr++)
			{
				if (*thisptr<128) { err = *thisptr;     pix8 |= b1; }
	             		else { err = *thisptr-255; pix8 |= w1; }

				if (bit==7) { *pp++ = pix8;  bit=pix8=0; }
 	       			else { pix8 <<= 1; bit++; }

				if (j<wide-1) thisptr[1] += ((err*7)/16);

				if (i<high-1)
				{
	  				nextptr[0] += ((err*5)/16);
	  				if (j>0) nextptr[-1] += ((err*3)/16);
	  				if (j<wide-1) nextptr[ 1] += (err/16);
				}
      			}

      			if (bit) *pp++ = pix8<<(7-bit);  /* write partial byte at end of line */
    		}
  	}

  	free(thisline);
	free(nextline);
}

/***********************************/
XImage *CreateXImage(theDisp, theVisual, pic, picType, width, hight, rMap, gMap, bMap)
Display *theDisp;
Visual *theVisual;
byte *pic, *rMap, *gMap, *bMap;
int picType;
int width, hight;
{
	byte *egampic;

	XImage *image = NULL;

	if (picType == PIC8) 
	{
		image = Pic8ToXImage(theDisp, theVisual, pic, (u_int) width, (u_int) hight,
					rMap, gMap, bMap); 
	}
	else if (picType == PIC24)
	{
		image = Pic24ToXImage(theDisp, theVisual, pic, (u_int) width, (u_int) hight );
	}

	return(image);
}

/***********************************/
XImage *Pic8ToXImage(theDisp, theVisual, pic8, wide, high, rmap, gmap, bmap)
Display *theDisp;
Visual *theVisual;
byte          *pic8, *rmap, *gmap, *bmap;
unsigned int   wide, high;
{
	/*
	 * this has to do the tricky bit of converting the data in 'pic8'
	 * into something usable for X.
	 *
	 */

	XImage *xim = NULL;
	byte  *imagedata = NULL;

	if (!pic8) return xim;  /* shouldn't happen */

	/* special case: 1-bit display */

	xim = XCreateImage(theDisp, theVisual, 1, XYBitmap, 0, NULL, 
		       wide, high, 32, 0);

/*	if (!xim) FatalError("couldn't create xim!");
*/
	imagedata = (byte *) malloc((size_t) (xim->bytes_per_line * high));

/*	if (!imagedata) FatalError("couldn't malloc imagedata");
*/
	xim->data = (char *) imagedata;
	floydDitherize(xim, pic8, PIC8, (int) wide, (int) high, rmap, gmap, bmap);

	return xim;
}

/***********************************/
XImage *Pic24ToXImage(theDisp, theVisual, pic24, wide, high)
Display *theDisp;
Visual *theVisual;
byte          *pic24;
unsigned int   wide, high;
{
	/*
	 * this has to do the none-to-simple bit of converting the data in 'pic24'
	 * into something usable by X.
	 *
	 * There are two major approaches:  if we're displaying on a TrueColor
	 * or DirectColor display, we've got all the colors we're going to need,
	 * and 'all we have to do' is convert 24-bit RGB pixels into whatever
	 * variation of RGB the X device in question wants.  No color allocation
	 * is involved.
	 *
	 * Alternately, if we're on a PseudoColor, GrayScale, StaticColor or 
	 * StaticGray display, we're going to continue to operate in an 8-bit 
	 * mode.  (In that by this point, a 3/3/2 standard colormap has been
	 * created for our use (though all 256 colors may not be unique...), and
	 * we're just going to display the 24-bit picture by dithering with those
	 * colors
	 *
	 */

	XImage *xim = NULL;
	byte  *imagedata = NULL;

	if (!pic24) return xim;  /* shouldn't happen */

	/* special case: 1-bit display.  Doesn't much matter *what* the visual is */

	xim = XCreateImage(theDisp, theVisual, 1, XYPixmap, 0, NULL, 
			        wide,  high, 32, 0);
/*	if (!xim) FatalError("couldn't create xim!");
*/
	imagedata = (byte *) malloc((size_t) (xim->bytes_per_line * high));

/*	if (!imagedata) FatalError("couldn't malloc imagedata");
*/
	xim->data = (char *) imagedata;

	floydDitherize(xim, pic24,PIC24, (int) wide, (int) high, NULL,NULL,NULL);

	return xim;
}

/*******************************/
int ReadImageFile(name, pinfo) 
     char    *name;
     GRaphe  *pinfo;
{
	int  i, ftype;
	char errstr[256];

	ftype = ReadFileType(name);

	if (ftype == RFT_ERROR)
	{
		sprintf(errstr, "Couldn't open file '%s'.", name);
		ErrorPopup(errstr);
		return 0;
	}
	else if(ftype == RFT_UNKNOWN)
	{
		sprintf(errstr, "Error:  File '%s' not in a recognized format.", name);
		ErrorPopup(errstr);
		return 0;
	}
	else
	{                         /* try to read it */
		i = ReadPicFile(name, ftype, pinfo, 1);

		if (!i || (i && (pinfo->w<=0 || pinfo->h<=0)))
		{
			if(i)
			{ 
				if (pinfo->pic) free(pinfo->pic);
			}

			sprintf(errstr, "Load file '%s' failed.", name);
			ErrorPopup(errstr);
			return 0;
		}
		/* got it! */
	}

	return 1;
}

int ReadPictureData(viewport, cp)
VIewportdata *viewport;
char *cp;
{
	GRaphe *curgra, *picPrt;

	curgra = (GRaphe *)malloc(sizeof(GRaphe));

	if(!ReadImageFile(cp, curgra))
	{
		free(curgra);
		return 1;
	}

	curgra->next = NULL;
	curgra->zoomNum = 0; 
	curgra->Image = NULL;
	curgra->fileName = strdup(cp);

	graphicsZoom(viewport, curgra, viewport->zoom);

	if(viewport->GraphRoot)
	{
		for(picPrt = viewport->GraphRoot; picPrt->next != NULL; picPrt= picPrt->next);
		picPrt->next = curgra;
	}
	else
	{
		viewport->GraphRoot = curgra;
	}

	return 0;
}

void InitialPictures(viewport)
VIewportdata *viewport;
{
	GRaphe *curgra, *picPrt, *lastPrt;

	if(viewport->GraphRoot)
	{
		for(lastPrt = NULL, picPrt = viewport->GraphRoot; picPrt != NULL; )
		{
			if(ReadImageFile(picPrt->fileName, picPrt))
			{
				graphicsZoom(viewport, picPrt, viewport->zoom);

				lastPrt = picPrt;
				picPrt= picPrt->next;
			}
			else
			{
				if(!lastPrt)
				{
					viewport->GraphRoot = picPrt->next;
				}
				else
				{
					lastPrt->next = picPrt->next;
				}

				curgra = picPrt;
				picPrt = picPrt->next; 
				free(curgra->fileName);
				free(curgra);
			}
		}
	}
}

void ReadPictureFilename(viewport, cp)
VIewportdata *viewport;
char *cp;
{
	GRaphe *curgra, *picPrt;

	curgra = (GRaphe *)malloc(sizeof(GRaphe));

	if(!curgra) return;

	curgra->fileName = strdup(cp);

        curgra->next = NULL;
        curgra->zoomNum = 0;
        curgra->Image = NULL;
	curgra->pic = NULL;

	if(viewport->GraphRoot)
	{
		for(picPrt = viewport->GraphRoot; picPrt->next != NULL; picPrt= picPrt->next);
		picPrt->next = curgra;
	}
	else
	{
		viewport->GraphRoot = curgra;
	}
}

GRaphe *SearchPicture(viewport, picname)
VIewportdata *viewport;
char *picname;
{
	GRaphe *picPrt;

	for(picPrt = viewport->GraphRoot; picPrt != NULL && strcmp(picname, picPrt->fileName) != 0; picPrt = picPrt->next);

	return(picPrt);
}

int UpdatePicture(viewport, picname)
VIewportdata *viewport;
char *picname;
{
	if(SearchPicture(viewport, picname)) return 0;
	else return(ReadPictureData(viewport, picname));
}

void WritePictureToFile(viewport, note)
VIewportdata *viewport;
char *note;
{
	GRaphe *picPrt;

	for(picPrt = viewport->GraphRoot; picPrt != NULL; picPrt= picPrt->next)
	{
		if(picPrt->used) fprintf(viewport->fp, "%s%s\n", note, picPrt->fileName);
	}
}

void ClearUpPic(viewport)
VIewportdata *viewport;
{
	GRaphe *picPrt;
	unsigned short *tp, *searchprt, chrmem;
	unsigned char *sp;
	int j;

	for(picPrt = viewport->GraphRoot; picPrt != NULL; picPrt= picPrt->next)
	{
		picPrt->used = False;
	}

	for(tp = viewport->EditFile; *tp != '\0'; tp++)
	{
		if(*tp == '@' && *(tp + 1) == '[' && *(tp + 2) == 'G')
		{
			tp += 3;

			for(searchprt = tp, j = 0; *searchprt != '\0' && *searchprt != ']' && j < MAXIDNUM; searchprt++, j++);
			if(j >= MAXIDNUM)
			{
				continue;
			}

			chrmem = *searchprt;
			*searchprt = '\0';

			sp = DBstrdup(tp);

			*searchprt = chrmem;
			tp = searchprt;

			if(sp)
			{
				picPrt = SearchPicture(viewport, sp);

				if(picPrt) 
				{
					picPrt->used = True;
				}
				free(sp);
			}
		}
	}
} 
		
int PictureCount(viewport)
VIewportdata *viewport;
{
	GRaphe *picPrt;
	int i;

	i = 0;

	for(picPrt = viewport->GraphRoot; picPrt != NULL; picPrt= picPrt->next, i++);

	return i;
}

void DeletePictureData(viewport)
VIewportdata *viewport;
{
	GRaphe *picPrt, *tmpPrt;

	for(picPrt = viewport->GraphRoot; picPrt != NULL;)
	{
		free(picPrt->fileName);
		free(picPrt->Image->data);
		XFree(picPrt->Image);
		free(picPrt->pic);
		tmpPrt = picPrt;
		picPrt= picPrt->next;
		free(tmpPrt);
	}

	viewport->GraphRoot = NULL;
}

void PictureToPS(viewport, graph)
VIewportdata *viewport;
GRaphe *graph;
{
	int i, j;
	int ptype = graph->picType, lwidth, rpix, gpix, bpix, w = graph->w, h = graph->h;
  	byte *inpix = graph->pic, *rmap = graph->r, *gmap = graph->g, *bmap = graph->b;

	fprintf(printset.psfile,"\ngsave ");

	if(!viewport->horOrder)
	{
		fprintf(printset.psfile,"0 -%d R ", graph->h * PICSCALE);
	}
	 
	fprintf(printset.psfile,"currentpoint translate %d %d scale %d %d 8 [%d 0 0 %d 0 %d] {currentfile %d string readhexstring pop} image", graph->w * PICSCALE, graph->h * PICSCALE, graph->w, graph->h, graph->w, -graph->h, graph->h, graph->w);

  for (i = 0; i < h ; i++) {
     	lwidth = 0;
     	putc('\n',printset.psfile);
 
        for (j = 0; j < w; j++) {
 
          if (ptype == PIC8) {
            rpix = rmap[*inpix];
            gpix = gmap[*inpix];
            bpix = bmap[*inpix];
          }
          else {  /* PIC24 */
            rpix = inpix[0];
            gpix = inpix[1];
            bpix = inpix[2];
          }
         
         fprintf(printset.psfile,"%02x", MONO(rpix,gpix,bpix));
         lwidth++;
 
         if (lwidth>40) { putc('\n',printset.psfile); lwidth = 0; }
 
         inpix += (ptype==PIC24) ? 3 : 1;
    }
 }
	if(viewport->horOrder)
	{
		fprintf(printset.psfile," G %d 0 R\n", graph->w * PICSCALE);
	}
	else
	{
		fprintf(printset.psfile," G 0 -%d R\n", graph->h * PICSCALE);
	}
}

