/* 
 * xvps.c - Postscript dialog box, file output functions
 *
 * callable functions:
 *
 *   CreatePSD(geom)           -  creates the psW window.  Doesn't map it.
 *   PSDialog()                -  maps psW
 *   PSCheckEvent(event)       -  called by event handler
 *   PSSaveParams(str,int)     -  tells PSDialog what to do when 'Ok' clicked
 *   PSResize()                -  called whenever ePic changes size
 *   LoadPS()                  -  attempts to load PS files using Ghostscript
 */

/***************************************************/
static void writePS(fp)
FILE *fp;
{
  int   i, j, q, err, rpix, gpix, bpix, nc, ptype;
  int   iw, ih, ox, oy, slen, lwidth, bits, colorps, w, h, pfree;
  double iwf, ihf;
  byte *inpix, *rmap, *gmap, *bmap;

  slen = bits = colorps = 0;
  
  /*** write PostScript header ***/

  if (colorType == F_BWDITHER) {   /* 1-bit dither code uses 'image' */
    int flipbw;

    /* set if color#0 is white */
    flipbw = (MONO(rmap[0],gmap[0],bmap[0]) > MONO(rmap[1],gmap[1],bmap[1]));

    fprintf(fp,"%% dimensions of data\n");
    fprintf(fp,"%d %d %d\n\n",w,h,bits);

    fprintf(fp,"%% mapping matrix\n");
    fprintf(fp,"[%d 0 0 %d 0 %d]\n\n", w, -h, h);

    fprintf(fp,"{currentfile pix readhexstring pop}\n");
    fprintf(fp,"image\n");

    /* write the actual image data */

    err = writeBWStip(fp, inpix, "", w, h, flipbw);
  }

  else {      /* all other formats */
    byte *rleline = (byte *) NULL;
    unsigned long outbytes = 0;

    /* if we're using color, make sure 'colorimage' is defined */
    if (colorps) psColorImage(fp);

    if (ptype==PIC8 && pscompCB.val) {  /* write cmap & rle-cmapped image fn */
      psColorMap(fp, colorps, nc, rmap, gmap, bmap);
      psRleCmapImage(fp, colorps);
    }

    fprintf(fp,"%d %d %d\t\t\t%% dimensions of data\n",w,h,bits);
    fprintf(fp,"[%d 0 0 %d 0 %d]\t\t%% mapping matrix\n", w, -h, h);

    if (ptype==PIC8 && pscompCB.val) fprintf(fp,"rlecmapimage\n");
    else {
      fprintf(fp,"{currentfile pix readhexstring pop}\n");
      if (colorps) fprintf(fp,"false 3 colorimage\n");
              else fprintf(fp,"image\n");
    }

    /* dump the image data to the file */
    err = 0;

    if (ptype==PIC8 && pscompCB.val) {
      rleline  = (byte *) malloc((size_t) w * 2);  /* worst-case */
      if (!rleline) FatalError("unable to malloc rleline in writePS()\n");
    }

    for (i=0; i<h && err != EOF; i++) {
      int rlen;
      lwidth = 0;
      putc('\n',fp);

      if ((i&0x1f) == 0) WaitCursor();

      if (ptype==PIC8 && pscompCB.val) { /* write rle-encoded cmapped image */
	rlen = rle_encode(inpix, rleline, w);
	inpix += w;
	outbytes += rlen;

	for (j=0; j<rlen && err != EOF; j++) {
	  err = fprintf(fp,"%02x", rleline[j]);
	  lwidth += 2;

	  if (lwidth>70) { putc('\n',fp); lwidth = 0; }
	}
      }

      else {  /* write non-rle raw (gray/rgb) image data */
	for (j=0; j<w && err != EOF; j++) {

	  if (ptype == PIC8) {
	    rpix = rmap[*inpix];
	    gpix = gmap[*inpix];
	    bpix = bmap[*inpix];
	  }
	  else {  /* PIC24 */
	    rpix = inpix[0];
	    gpix = inpix[1];
	    bpix = inpix[2];
	  }
	  
	  if (colorps) { 
	    err = fprintf(fp,"%02x%02x%02x",rpix,gpix,bpix);
	    lwidth+=6;
	  }
      
	  else {  /* greyscale */
	    err = fprintf(fp,"%02x", MONO(rpix,gpix,bpix));
	    lwidth+=2;
	  }

	  if (lwidth>70) { putc('\n',fp); lwidth = 0; }

	  inpix += (ptype==PIC24) ? 3 : 1;
	}
      }
    }

    if (ptype==PIC8 && pscompCB.val) {
      free(rleline);
      fprintf(fp,"\n\n");
      fprintf(fp,"%%\n");
      fprintf(fp,"%% Compression made this file %.2f%% %s\n",
	      100.0 * ((double) outbytes) / 
	      ((double) eWIDE * eHIGH * ((colorps) ? 3 : 1)),
	      "of the uncompressed size.");
      fprintf(fp,"%%\n");
    }
  }


  fprintf(fp,"\n\nshowpage\n\n");

  fprintf(fp,"%% stop using temporary dictionary\n");
  fprintf(fp,"end\n\n");

  fprintf(fp,"%% restore original state\n");
  fprintf(fp,"origstate restore\n\n");
  fprintf(fp,"%%%%Trailer\n");

  if (pfree) free(inpix);

  if (CloseOutFile(fp, filename, (err==EOF)) == 0) {
    DirBox(0);
  }

  SetCursors(-1);
}


/**********************************************/
static int rle_encode(scanline, rleline, wide)
     byte *scanline, *rleline;
     int wide;
{
  /* generates a rle-compressed version of the scan line.
   * rle is encoded as such:
   *    <count> <value>                      # 'run' of count+1 equal pixels
   *    <count | 0x80> <count+1 data bytes>  # count+1 non-equal pixels
   *
   * count can range between 0 and 127
   *
   * returns length of the rleline vector
   */
  
  int  i, j, blocklen, isrun, rlen;
  byte block[256], pix;
  
  blocklen = isrun = rlen = 0;

  for (i=0; i<wide; i++) {
    /* there are 5 possible states:
     *   0: block empty.
     *   1: block not empty, block is  a run, current pix == previous pix
     *   2: block not empty, block is  a run, current pix != previous pix
     *   3: block not empty, block not a run, current pix == previous pix
     *   4: block not empty, block not a run, current pix != previous pix
     */

    pix = scanline[i];

    if (!blocklen) {                    /* case 0:  empty */
      block[blocklen++] = pix;
      isrun = 1;
    }

    else if (isrun) {
      if (pix == block[blocklen-1]) {   /* case 1:  isrun, prev==cur */
	block[blocklen++] = pix;
      }
      else {                            /* case 2:  isrun, prev!=cur */
	if (blocklen>1) {               /*   we have a run block to flush */
	  rleline[rlen++] = blocklen-1;
	  rleline[rlen++] = block[0];
	  block[0] = pix;               /*   start new run block with pix */
	  blocklen = 1;
	}
	else {
	  isrun = 0;                    /*   blocklen<=1, turn into non-run */
	  block[blocklen++] = pix;
	}
      }
    }
	
    else {   /* not a run */
      if (pix == block[blocklen-1]) {   /* case 3:  non-run, prev==cur */
	if (blocklen>1) {               /*  have a non-run block to flush */
	  rleline[rlen++] = (blocklen-1) | 0x80;
	  for (j=0; j<blocklen; j++)
	    rleline[rlen++] = block[j];

	  block[0] = pix;               /*  start new run block with pix */
	  blocklen = isrun = 1;
	}
	else {
	  isrun = 1;                    /*  blocklen<=1 turn into a run */
	  block[blocklen++] = pix;
	}
      }
      else {                            /* case 4:  non-run, prev!=cur */
	block[blocklen++] = pix;
      }
    }

    if (blocklen == 128) {   /* max block length.  flush */
      if (isrun) {
	rleline[rlen++] = blocklen-1;
	rleline[rlen++] = block[0];
      }

      else {
	rleline[rlen++] = (blocklen-1) | 0x80;
	for (j=0; j<blocklen; j++) 
	  rleline[rlen++] = block[j];
      }

      blocklen = 0;
    }
  }

  if (blocklen) {   /* flush last block */
    if (isrun) {
      rleline[rlen++] = blocklen-1;
      rleline[rlen++] = block[0];
    }

    else {
      rleline[rlen++] = (blocklen-1) | 0x80;
      for (j=0; j<blocklen; j++) 
	rleline[rlen++] = block[j];
    }
  }

  return rlen;
}
	  
	    
/**********************************************/
static void psColorImage(fp)
FILE *fp;
{
  /* spits out code that checks if the PostScript device in question
     knows about the 'colorimage' operator.  If it doesn't, it defines
     'colorimage' in terms of image (ie, generates a greyscale image from
     RGB data) */


  fprintf(fp,"%% define 'colorimage' if it isn't defined\n");
  fprintf(fp,"%%   ('colortogray' and 'mergeprocs' come from xwd2ps\n");
  fprintf(fp,"%%     via xgrab)\n");
  fprintf(fp,"/colorimage where   %% do we know about 'colorimage'?\n");
  fprintf(fp,"  { pop }           %% yes: pop off the 'dict' returned\n");
  fprintf(fp,"  {                 %% no:  define one\n");
  fprintf(fp,"    /colortogray {  %% define an RGB->I function\n");
  fprintf(fp,"      /rgbdata exch store    %% call input 'rgbdata'\n");
  fprintf(fp,"      rgbdata length 3 idiv\n");
  fprintf(fp,"      /npixls exch store\n");
  fprintf(fp,"      /rgbindx 0 store\n");
  fprintf(fp,"      0 1 npixls 1 sub {\n");
  fprintf(fp,"        grays exch\n");
  fprintf(fp,"        rgbdata rgbindx       get 20 mul    %% Red\n");
  fprintf(fp,"        rgbdata rgbindx 1 add get 32 mul    %% Green\n");
  fprintf(fp,"        rgbdata rgbindx 2 add get 12 mul    %% Blue\n");
  fprintf(fp,"        add add 64 idiv      %% I = .5G + .31R + .18B\n");
  fprintf(fp,"        put\n");
  fprintf(fp,"        /rgbindx rgbindx 3 add store\n");
  fprintf(fp,"      } for\n");
  fprintf(fp,"      grays 0 npixls getinterval\n");
  fprintf(fp,"    } bind def\n\n");

  fprintf(fp,"    %% Utility procedure for colorimage operator.\n");
  fprintf(fp,"    %% This procedure takes two procedures off the\n");
  fprintf(fp,"    %% stack and merges them into a single procedure.\n\n");
  
  fprintf(fp,"    /mergeprocs { %% def\n");
  fprintf(fp,"      dup length\n");
  fprintf(fp,"      3 -1 roll\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      length\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      5 1 roll\n");
  fprintf(fp,"      3 -1 roll\n");
  fprintf(fp,"      add\n");
  fprintf(fp,"      array cvx\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      3 -1 roll\n");
  fprintf(fp,"      0 exch\n");
  fprintf(fp,"      putinterval\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      4 2 roll\n");
  fprintf(fp,"      putinterval\n");
  fprintf(fp,"    } bind def\n\n");

  fprintf(fp,"    /colorimage { %% def\n");
  fprintf(fp,"      pop pop     %% remove 'false 3' operands\n");
  fprintf(fp,"      {colortogray} mergeprocs\n");
  fprintf(fp,"      image\n");
  fprintf(fp,"    } bind def\n");
  fprintf(fp,"  } ifelse          %% end of 'false' case\n");
  fprintf(fp,"\n\n\n");
}


/**********************************************/
static void psColorMap(fp, color, nc, rmap, gmap, bmap)
     FILE *fp;
     int color, nc;
     byte *rmap, *gmap, *bmap;
{
  /* spits out code for the colormap of the following image
     if !color, it spits out a mono-ized graymap */

  int i;

  fprintf(fp,"%% define the colormap\n");
  fprintf(fp,"/cmap %d string def\n\n\n", nc * ((color) ? 3 : 1));

  fprintf(fp,"%% load up the colormap\n");
  fprintf(fp,"currentfile cmap readhexstring\n");

  for (i=0; i<nc; i++) {
    if (color) fprintf(fp,"%02x%02x%02x ", rmap[i],gmap[i],bmap[i]);
    else fprintf(fp,"%02x ", MONO(rmap[i],gmap[i],bmap[i]));
    
    if ((i%10) == 9) fprintf(fp,"\n");
  }
  if (i%10) fprintf(fp,"\n");
  fprintf(fp,"pop pop   %% lose return values from readhexstring\n\n\n");
		 
}


/**********************************************/
static void psRleCmapImage(fp, color)
FILE *fp;
int   color;
{
  /* spits out code that defines the 'rlecmapimage' operator */

  fprintf(fp,"%% rlecmapimage expects to have 'w h bits matrix' on stack\n");
  fprintf(fp,"/rlecmapimage {\n");
  fprintf(fp,"  /buffer 1 string def\n");
  fprintf(fp,"  /rgbval 3 string def\n");
  fprintf(fp,"  /block  384 string def\n\n");

  fprintf(fp,"  %% proc to read a block from file, and return RGB data\n");
  fprintf(fp,"  { currentfile buffer readhexstring pop\n");
  fprintf(fp,"    /bcount exch 0 get store\n");
  fprintf(fp,"    bcount 128 ge\n");
  fprintf(fp,"    {  %% it's a non-run block\n");
  fprintf(fp,"      0 1 bcount 128 sub\n");
  fprintf(fp,"      { currentfile buffer readhexstring pop pop\n\n");

  if (color) {
    fprintf(fp,"        %% look up value in color map\n");
    fprintf(fp,"%s/rgbval cmap buffer 0 get 3 mul 3 getinterval store\n\n",
	    "        ");
    fprintf(fp,"        %% and put it in position i*3 in block\n");
    fprintf(fp,"        block exch 3 mul rgbval putinterval\n");
    fprintf(fp,"      } for\n");
    fprintf(fp,"      block  0  bcount 127 sub 3 mul  getinterval\n");
    fprintf(fp,"    }\n\n");
  }
  else {
    fprintf(fp,"        %% look up value in gray map\n");
    fprintf(fp,"%s/rgbval cmap buffer 0 get 1 getinterval store\n\n",
	    "        ");
    fprintf(fp,"        %% and put it in position i in block\n");
    fprintf(fp,"        block exch rgbval putinterval\n");
    fprintf(fp,"      } for\n");
    fprintf(fp,"      block  0  bcount 127 sub  getinterval\n");
    fprintf(fp,"    }\n\n");
  }

  fprintf(fp,"    { %% else it's a run block\n");
  fprintf(fp,"      currentfile buffer readhexstring pop pop\n\n");

  if (color) {
    fprintf(fp,"      %% look up value in colormap\n");
    fprintf(fp,"%s/rgbval cmap buffer 0 get 3 mul 3 getinterval store\n\n",
	    "      ");
    fprintf(fp,"%s0 1 bcount { block exch 3 mul rgbval putinterval } for\n\n",
	    "      ");
    fprintf(fp,"      block 0 bcount 1 add 3 mul getinterval\n");
  }
  else {
    fprintf(fp,"      %% look up value in graymap\n");
    fprintf(fp,"      /rgbval cmap buffer 0 get 1 getinterval store\n\n");
    fprintf(fp,"      0 1 bcount { block exch rgbval putinterval } for\n\n");
    fprintf(fp,"      block 0 bcount 1 add getinterval\n");
  }

  fprintf(fp,"    } ifelse\n");
  fprintf(fp,"  } %% end of proc\n");

  if (color) fprintf(fp,"  false 3 colorimage\n");
        else fprintf(fp,"  image\n");

  fprintf(fp,"} bind def\n\n\n");
}



/**********************************************/
static void epsPreview(fp, pic, ptype, colorType, w, h, rmap,gmap,bmap, 
		       landscape)
     FILE *fp;
     byte *pic;
     int   ptype, colorType;
     int   w, h, landscape;
     byte *rmap, *gmap, *bmap;
{
  byte *prev;
  int flipbw;


  if (landscape) {  /* generate a rotated version of the pic */
    int bperpix;
    byte *lpic;

    bperpix = (ptype == PIC8) ? 1 : 3;
    lpic = (byte *) malloc((size_t) w * h * bperpix);
    if (!lpic) FatalError("can't alloc mem to rotate image in epsPreview");

    xvbcopy((char *) pic, (char *) lpic, (size_t) w * h * bperpix);
    RotatePic(lpic, ptype, &w, &h, 0);
    pic = lpic;
  }
    

  /* put in an EPSI preview */
  
  if (colorType != F_BWDITHER) { /* have to generate a preview */
    prev = FSDither(pic, ptype, w, h, rmap,gmap,bmap, 0, 1);

    if (!prev) {
      fprintf(stderr,"Unable to malloc in epsPreview\n");
      return;
    }

    flipbw = 0;
  }
  else {
    prev = pic;
    /* set if color#0 is white */
    flipbw = (MONO(rmap[0],gmap[0],bmap[0]) > MONO(rmap[1],gmap[1],bmap[1]));
  }

 
  fprintf(fp,"%%%%BeginPreview: %d %d %d %d\n", w, h, 1, 
	  (w/(72*4) + 1) * h);

  writeBWStip(fp, prev, "% ", w, h, !flipbw);

  fprintf(fp,"%%%%EndPreview\n");

  if (colorType != F_BWDITHER) free(prev);
  if (landscape) free(pic);  /* lpic, actually... */
}


/***********************************/
static int writeBWStip(fp, pic, prompt, w, h, flipbw)
     FILE *fp;
     byte *pic;
     char *prompt;
     int  w, h, flipbw;
{
  /* write the given 'pic' (B/W stippled, 1 byte per pixel, 0=blk,1=wht) 
     out as hexadecimal, max of 72 hex chars per line.

     if 'flipbw', then 0=white, 1=black

     returns '0' if everythings fine, 'EOF' if writing failed */

  int err, i, j, lwidth;
  byte outbyte, bitnum, bit;

  err = 0;

  for (i=0; i<h && err != EOF; i++) {
    fprintf(fp, "%s", prompt);

    outbyte = bitnum = lwidth = 0;

    if ((i&0x3f) == 0) WaitCursor();

    for (j=0; j<w && err != EOF; j++) {
      bit = *pic;
      outbyte = (outbyte<<1) | ((bit)&0x01);
      bitnum++;

      if (bitnum==8) {
	if (flipbw) outbyte = ~outbyte & 0xff;
	err = fprintf(fp,"%02x",outbyte);
	lwidth+=2;
	outbyte = bitnum = 0;
      }

      if (lwidth>=72 && j+1<w) { fprintf(fp, "\n%s", prompt); lwidth = 0; }
      pic++;
    }

    if (bitnum) {   /* few bits left over... */
      for ( ; bitnum<8; bitnum++) outbyte <<= 1;
      if (flipbw) outbyte = ~outbyte & 0xff;
      err = fprintf(fp,"%02x",outbyte);
      lwidth+=2;
    }

    fprintf(fp, "\n");
  }

  return err;
}

